INDICES = {
    "dl19": "msmarco-v1-passage",
    "dl20": "msmarco-v1-passage",
    "dl21": "msmarco-v2-passage",
    "dl22": "msmarco-v2-passage",
    "covid": "beir-v1.0.0-trec-covid.flat",
    "arguana": "beir-v1.0.0-arguana.flat",
    "touche": "beir-v1.0.0-webis-touche2020.flat",
    "news": "beir-v1.0.0-trec-news.flat",
    "scifact": "beir-v1.0.0-scifact.flat",
    "fiqa": "beir-v1.0.0-fiqa.flat",
    "scidocs": "beir-v1.0.0-scidocs.flat",
    "nfc": "beir-v1.0.0-nfcorpus.flat",
    "quora": "beir-v1.0.0-quora.flat",
    "dbpedia": "beir-v1.0.0-dbpedia-entity.flat",
    "fever": "beir-v1.0.0-fever.flat",
    "robust04": "beir-v1.0.0-robust04.flat",
    "signal": "beir-v1.0.0-signal1m.flat",
    "mrtydi-ar": "mrtydi-v1.1-arabic",
    "mrtydi-bn": "mrtydi-v1.1-bengali",
    "mrtydi-fi": "mrtydi-v1.1-finnish",
    "mrtydi-id": "mrtydi-v1.1-indonesian",
    "mrtydi-ja": "mrtydi-v1.1-japanese",
    "mrtydi-ko": "mrtydi-v1.1-korean",
    "mrtydi-ru": "mrtydi-v1.1-russian",
    "mrtydi-sw": "mrtydi-v1.1-swahili",
    "mrtydi-te": "mrtydi-v1.1-telugu",
    "mrtydi-th": "mrtydi-v1.1-thai",
}
